# -*- coding: iso-8859-1 -*-
"""
    MoinMoin - metadata search action
    
    This is the backend for the search.  Searches pages and prints results.
    
    Original version Largely refactored from "fullsearch" by Jürgen Hermann, 
    but much has changed since then.

    Changelog:
        0.      Initial version
        0.1     Changed querystr format to meta:"key = value"
    
    @version: 0.1
    @copyright: (c) 2005 by Charlie DeTar
    @license: Gnu GPL, see COPYING for details.
"""

from MoinMoin.Page import Page
from MoinMoin import wikiutil, search
from MoinMoin.formatter.text_html import Formatter
from MoinMoin.util import MoinMoinNoFooter

import re

def execute(pagename, request, fieldname='value', titlesearch=0):
    """ Copied verbatim from fullsearch.py action, with the singular 
    change of calling a Meta-capable query parser object.
    """
    _ = request.getText

    # Get titlesearch. It either a hidden field in a search macro form
    # using '0' or '1', or a submit button in the search box, in which
    # case its a localized 'Titles' string.
    try:
        titlesearch = int(request.form.get('titlesearch', [0])[0])
    except ValueError:
        # "Titles" button was clicked in the standard search box.
        titlesearch = 1

    # context is relevant only for full search
    if titlesearch:
        context = 0
    else:
        context = int(request.form.get('context', [0])[0])

    # Get other form parameters
    needle = request.form.get(fieldname, [''])[0]
    case = int(request.form.get('case', [0])[0])
    regex = int(request.form.get('regex', [0])[0]) # no interface currently

    max_context = 1 # only show first `max_context` contexts XXX still unused

    # check for sensible search term
    striped = needle.strip()
    if len(striped) == 0:
        err = _('Please use a more selective search term instead '
                'of {{{"%s"}}}') % needle
        # send http headers
        request.http_headers()
        Page(request, pagename).send_page(request, msg=err)
        return

    # search the pages
    query = MetaCapableQueryParser(case=case, regex=regex,
                               titlesearch=titlesearch).parse_query(needle)
    results = search.searchPages(request, query)

    # directly show a single hit
    # XXX won't work with attachment search
    # improve if we have one...
    if len(results.hits) == 1:
        page = Page(request, results.hits[0].page_name)
        # TODO: remove escape=0 in 1.4
        url = page.url(request, querystr={'highlight': query.highlight_re()},
                       escape=0)
        request.http_redirect(url)
        raise MoinMoinNoFooter

    # send http headers
    request.http_headers()

    # This action generate data using the user language
    request.setContentLanguage(request.lang)

    # Setup for type of search
    if titlesearch:
        title = _('Title Search: "%s"')
        results.sortByPagename()
    else:
        title = _('Full Text Search: "%s"')
        results.sortByWeight()

    wikiutil.send_title(request, title % needle, form=request.form,
                        pagename=pagename)

    # Start content (important for RTL support)
    formatter = Formatter(request)
    request.write(formatter.startContent("content"))

    # First search stats
    request.write(results.stats(request, formatter))

    # Then search results
    info = not titlesearch
    if context:
        output = results.pageListWithContext(request, formatter, info=info,
                                             context=context)
    else:
        output = results.pageList(request, formatter, info=info)
    request.write(output)

    # End content and send footer
    request.write(formatter.endContent())
    wikiutil.send_footer(request, pagename, editable=0, showactions=0,
                         form=request.form)

class MetaSearch(search.BaseExpression):

    # This regular expression causes a recursion exceeded error.  Not
    # sure why...
    #
    # search_area = r'{{{\s*#!meta((?:(.|\n)(|!}}}))*(.|\n))}}}'
    # 
    # So I changed it to this, I hope this doesn't change the
    # functionality:
    #
    search_area = r'{{{\s*#!meta(((.|\n)*?))}}}'


    def __init__(self, text = None, search_area = None, case=0):
        if (search_area != None):
            self.search_area = search_area
        self.area_re = re.compile(self.search_area)
        if text:
            self.metaterms = self._parseQueryString(text)
        else:
            self.metaterms = []
        self.case = case
        self.negated = 0

    def costs(self):
        ''' costs more than full text search '''
        return 10001
    
    def search(self, page):
        matches = []
                
        # Search in page 
        found = self.matchesText(page)
        if found:
            matches.append(search.TextMatch(0, 100))

        # Decide what to do with the results.
        if ((self.negated and matches) or
            (not self.negated and not matches)):
            return None
        elif matches:
            return matches
        else:
            return [search.Match()]
    
    def _parseQueryString(self, querystr):
        """ Query strings of this form:
        key = value ; key2 = value2 ; key3 = value3
        """
        parts = querystr.split(';')
        terms = []
        for part in parts:
            try:
                key, value = part.split('=')
            except ValueError:
                key = part
                value = ''
            terms.append( (key.strip(), value.strip()) )
        return terms
    
    def getRelevantPortion(self, text):
        """ Compile all meta blocks from page. """
        searchText = ""
        while True:
            match = self.area_re.search(text)
            if match:
                searchText += match.group(1)
                text = text[match.end(1):]
            else:
                break

        return searchText
        
    def matchesText(self, page):
        """ Return true if the keys/values in self.terms match
        the meta data block in text parameter."""
        # Search for a given needle in the page first:
        searchText = self.getRelevantPortion( page.get_raw_body() )

        # set up re flags:
        if self.case:
            flags = re.U
        else:
            flags = re.U | re.I

        # search in meta blocks for key/value pairs.
        if searchText:
            lines = searchText.split('\n')
            
            # default to found, if there are no key/value pairs...
            found = 1
            for key,value in self.metaterms:
                # default to 0 if there are key/value pairs
                found = 0
                for line in lines:
                    if len(line) > 0 and line[0] != '#':
                        try:
                            keystr,valuestr = line.split(':')
                        except ValueError:
                            keystr = line
                            valuestr = ""
                        if key:
                            if not keystr:
                                continue
                            reg = "(^|[\s;/])"+key+"([\s;:/]|$)"
                            search_re = re.compile(reg, flags) 

                            if not search_re.search(keystr):
                                continue

                        if value:
                            if not valuestr:
                                continue
                            reg = "(^|[\s;:])"+value+"([\s;]|$)"
                            search_re = re.compile(reg, flags) 

                            if not search_re.search(valuestr):
                                continue
                            
                        if key or value:
                            found = 1
                if not found:
                    break
            if found:
                return searchText
        return False

class MetaCapableQueryParser(search.QueryParser):
    
    def _single_term(self):
        ''' copied from parent and modified in 
        three places (marked below) 
        '''
        regex = (r'(?P<NEG>-?)\s*(' +              # leading '-'
                 r'(?P<OPS>\(|\)|(or\b(?!$)))|' +  # or, (, )
                 r'(?P<MOD>(\w+:)*)' +
                 r'(?P<TERM>("[^"]+")|' +
                  r"('[^']+')|(\S+)))")             # search word itself
        self._query = self._query.strip()
        match = re.match(regex, self._query, re.U)
        if not match:
            return None
        self._query = self._query[match.end():]
        ops = match.group("OPS")
        if ops == '(':
            result = self._or_expression()
            if match.group("NEG"): restult.negate()
            return result
        elif ops == ')':
            return None
        elif ops == 'or':
            return None
        modifiers = match.group('MOD').split(":")[:-1]
        text = match.group('TERM')
        if ((text[0] == text[-1] == '"') or
            (text[0] == text[-1] == "'")): text = text[1:-1]

        title_search = self.titlesearch
        regex = self.regex
        case = self.case
        linkto = 0
        # modification
        meta = False

        for m in modifiers:
            if "title".startswith(m):
                title_search = True
            elif "regex".startswith(m):
                regex = True
            elif "case".startswith(m):
                case = True
            elif "linkto".startswith(m):
                linkto = True
            # modification
            elif "meta".startswith(m):
                meta = True

        # modification: added search. namespace to objects,
        # added MetaSearch option.
        if linkto:
            obj = search.LinkSearch(text, use_re=regex, case=case)
        elif title_search:
            obj = search.TitleSearch(text, use_re=regex, case=case)
        elif meta:
            obj = MetaSearch(text, case=case)
        else:
            obj = search.TextSearch(text, use_re=regex, case=case)

        if match.group("NEG"):
            obj.negate()
        return obj

