"""
MoinMoin meta-data parser

Logs and formats keys and values of meta-data.

Syntax:
    * newlines separate entries.  
    * Colons (:) split key/value pairs.
    * Either the key or the value may be ommitted.
    * Semicolons are multipliers:  
        one; two; three: value
      expands to:
        one: value
        two: value
        three: value
      This works for keys or values or both.

    * Slashes in keys demarcate subkeys:
        rootkey/subkey/subsubkey : value

    * Hash (#) demarcates a special command or comment.
        possible values for special commands:
            #hidden
            #list
            #table (default)
            #plain
            
            #class <css classname>
            #title <string or none>

    Every key/value is a link to a search page for their
    respective values.  (requires metasearch.py action to be installed).
    
    Example:
    
    {{{#!meta
        #list
        key: value 
        one key: many, values
        many; keys; here: many; values
        key without values
        : value without keys
    }}}

Changelog:
    0.0     initial version
    0.1     changed querystring format to meta:"key = value", 
            changed multiplicative operator to ; from ,.

@version: 0.1 
@copyright: 2005 by Charlie DeTar <charlie@theendmusic.org>
@license: GNU GPL, see COPYING for details.
"""


import re,urllib 

from MoinMoin import wikiutil

class MetaParser: 
    # raw html heading for categories div:
    heading = "Meta Data"
    # meta browser wiki page:
    browser = "BrowseMeta"
    # css class for categories div:
    cssClass = "metadata"
    # css class for meta keys 
    keyClass = "metakey"
    # css class for meta values 
    valueClass = "metavalue"
    # Set to true to display values as meta-search links
    valuesAreSearchLinks = 0
    # action for links from meta data
    action = "metasearch"

    # default display
    display = "table"

    null = '%00'
    # characters disallowed for keys and values
    disallowed = re.compile('["]')

    isLinkRe = re.compile('https?://[^\s]+')
    isMailRe = re.compile('[-a-zA-Z0-9_.]+@[-a-zA-Z0-9_.]+')
    
    def __init__(self, raw, request, **kw):
        self.raw = raw
        self.request = request
        self.form = request.form
        self._ = request.getText
        if self.raw:
            self.parseMetadata(self.raw)

        # Set up default parser for displaying any plain text
        Parser = wikiutil.importWikiPlugin( self.request.cfg, "parser", 
                            self.request.cfg.default_markup, "Parser") 
        if Parser is None:
            del Parser
            from MoinMoin.parser.plain import Parser
        self.DefaultParser = Parser
        
        
    def parseMetadata(self, raw):
        # lines that are parsed as commands (or ignored)
        command_re = re.compile('^#([^\s]*)(?:\s(.*))?$') 

        lines = raw.split('\n')
        metadata = []
        for line in lines:
            commandMatch = command_re.match(line)
            if commandMatch:
                command = commandMatch.group(1)
                displayTypes = ['hidden','table','plain','list']
                try:
                    which = displayTypes.index(command)
                except ValueError:
                    which = -1
                if which != -1:
                    self.display = displayTypes[which]
                elif command == 'title' or command == 'heading':
                    self.heading = commandMatch.group(2)
                elif command == 'class':
                    self.cssClass = commandMatch.group(2)
                continue       

            splitpoint = line.find(':')
            if splitpoint == -1:
                keys = [self.filter(key) for key in line.split(';')]
                values = [""]
            else:
                keys = [self.filter(key) for key in line[0 : splitpoint].split(';')]
                values = [self.filter(value) for value in line[splitpoint + 1 : ].split(';')]

            for key in keys:
                if key or values[0] or len(values) > 1:
                    metadata.append((key, values))
        
        self.metadata = metadata
        return metadata

    def filter(self, string):
        return self.disallowed.sub('',string).strip()

    def format(self, formatter):
        """ Send the text. """
        _ = self._
        if self.display == "hidden":
            return

        write = self.request.write
        f = formatter
        dest = self.request.page.page_name
        null = self.null
            
        write(f.open('div', attr = {'id': self.cssClass }))
        
        #########  Table  ###############################
        
        if self.display == "table":
            write(f.table(1))
            if self.heading:
                
                write(f.open('th', attr = {'colspan':'2'} ))
                write(self.formatHeading(f))
                write(f.close('th'))
            
            for key,values in self.metadata:
                write(f.table_row(1))
                
                write(f.table_cell(1, attrs = {'class': self.keyClass}))
                write(self.buildLinks(f, key))
                write(f.table_cell(0))
                
                write(f.table_cell(1, attrs = {'class': self.valueClass}))
                # metadata values link to both key and value
                for value in values:
                    if self.valuesAreSearchLinks:
                        write(self.buildLinks(f, (key,value)))
                        write(f.linebreak(0))
                    else:
                        self.DefaultParser(value, self.request).format(f)
                write(f.table_cell(0))
                
                write(f.table_row(0))
            
            write(f.table(0))
        
        #########  List  ###############################

        elif self.display == "ul" or self.display == "list":
            if self.heading:
                write(self.formatHeading(f))

            write(f.bullet_list(1))
            for key, values in self.metadata:
                write(f.listitem(1, css_class = self.keyClass))
                write(self.buildLinks(f, key))
                if values[0] or len(values) > 1:
                    write(f.bullet_list(1))
                    for value in values:
                        write(f.listitem(1, css_class = self.valueClass))
                        if self.valuesAreSearchLinks:
                            write(self.buildLinks(f, (key, value) ))
                        else:
                            self.DefaultParser(value, self.request).format(f)
                        write(f.listitem(0))
                    write(f.bullet_list(0))
                write(f.listitem(0))
            write(f.bullet_list(0))
        
        #########  Plain  ###############################

        else: #if self.display == "plain":
            if self.heading:
                write(self.formatHeading(f))
            write(f.linebreak(0))
            for key, values in self.metadata:
                write(f.open('span', attr = { 'class': self.keyClass }))
                write(self.buildLinks(f, key))
                write(f.close('span'))
                if values[0] or len(values) > 1:
                    write(f.text(': '))
                    write(f.open('span', attr = { 'class': self.valueClass }))
                    for i,value in enumerate(values):
                        if self.valuesAreSearchLinks:
                            write(self.buildLinks(f, (key, value) ))
                        else:
                            self.DefaultParser(value, self.request).format(f)
                        if i < len(values) - 1:
                            write(f.text(', '))
                    write(f.close('span'))
        
        write(f.close('div'))
                                
#TODO:  implelment meaningful RDF.
            # RDF 
#           head = """
#               <!--
#               <?xml version="1.0"  encoding="iso-8859-1"?>
#               <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#">
#                   <rdf:Description rdf:about="%s">\n""" % self.request.getBaseURL()
            
#           fields = ""
#           for key,valuess in self.metadata:
#               key = _(key)
#               for value in values:
#                   value = _(value)
#                   fields += '\t\t\t<%s>%s</%s>\n' % (key,value,key)
#           foot = """\t\t</ rdf:Description>
#               </rdf:RDF>    
#               -->
#               """
#           rdf = head + fields + foot
#           write( rdf )
            
    def formatHeading(self, formatter):
        out = ""
#        out += formatter.pagelink(1, self.browser)
        out += formatter.text(self.heading)
#        out += formatter.pagelink(0)
        return out


    def isLink(self, word):
        return self.isLinkRe.match(word)

    def isMail(self, word):
        return self.isMailRe.match(word)

    def buildLinks(self, formatter, data):
        if len(data) == 2:
            type = 'value'
            key,value = data
            if self.isLink( value ):
                return self.formatLink(formatter, value, value, absolute = 1)
            if self.isMail( value.strip() ):
                return self.formatLink(formatter, value, "mailto:"+value, absolute = 1)
        else:
            type = 'key'
            key = data

        out = ""
        # For key links, we link to BrowseMeta 
        # with #key (anchor)
        if type == 'key':
            links = self.splitSubKeys(key)
            for i,(text,params) in enumerate(links):
                href = self.buildAnchorStr(params)
                out += self.formatLink(formatter, text, href)
                if i < len(links) - 1:
                    out += formatter.text("/")
        # For value links, we link to search
        # for key/value pair
        else:
            href = self.buildQueryStr(data)
            text = value
            out += self.formatLink(formatter, text, href)
        return out

    def formatLink(self, formatter, text, href, absolute = 0):
        """ outputs a link """
        out = ""
        if text and text != self.null:
            if absolute == 0:
                out += wikiutil.link_tag(
                            request=self.request, 
                            params=href,
                            formatter=formatter, 
                            on=1
                    )
                out += text
                out += wikiutil.link_tag(formatter,href,on=0)
            else:
                out += formatter.url(1, url=href)
                out += text
                out += formatter.url(0)
        return out

    def buildQueryStr(self, data ):
        """ build a query string.

        Format:  value=meta:"key = value"
        """
        q = u'''meta:"%s=%s" '''
        str = u"%s?action=%s&value=" % (self.request.page.page_name, self.action)
        # build query string
        str += urllib.quote(q % data)
        return str

    def buildAnchorStr(self, data ):
        str = "%s#%s" % (self.browser, data[0])
        return str

    def keyValueDictToPairs(self, data):
        # extract key/value pairs from data:
        fields = [['',''] for i in xrange(len(data)/2)]
        count = 0
        for type, field in data.iteritems():
            if type == 'key':
                fields[count/2][0] = field
            elif type == 'value':
                fields[count/2][1] = field
            count += 1
        return fields

    def splitSubKeys(self, keys, parents = None, links = None, demarcation = "/"):
        """ Split subkeys by the given demarcation.  
        Each key is given its own subkeys include their parents in the link.

        Returns a list of (key, parameters) tuples, where the parametrs are a
        {'key': key, 'value': value} dict.
        """
        if links == None:
            links = []
        if parents == None:
            parents = []

        subkey = keys.find(demarcation)
        if subkey == -1:
            key = keys
            reversed = parents[:]
            reversed.reverse()
            fullkey = ""
            for parent in parents:
                fullkey += parent + demarcation 
            fullkey += key
            params = ( fullkey, '') 
            links.append( (key, params) )
        else:
            key = keys[:subkey]
            self.splitSubKeys( key, parents, links )
            children = keys[subkey + 1:]
            subparents = parents[:]
            subparents.append( key )
            # recurse
            self.splitSubKeys( children, subparents, links )
        return links

Parser = MetaParser           

